# xmltypes.py - base XML classes, part of osa.
# Copyright 2013 Sergey Bozhenkov, boz at ipp.mpg.de
# Licensed under LGPLv3 or later, see the COPYING file.

"""
    Python classes corresponding to XML schema.
"""
from . import xmlnamespace
from . import xmlparser
from decimal import Decimal
from datetime import date, datetime, timedelta
import xml.etree.cElementTree as etree
import base64
import sys
if sys.version_info[0] > 2:
    unicode = str


def get_local_type(xmltype):
    """
        Simplifies types names, e.g. XMLInteger is
        presented as int.

        This is used for nice printing only.
    """
    if xmltype == "XMLBoolean":
        return 'bool'
    elif xmltype == "XMLDecimal":
        return 'decimal'
    elif xmltype == "XMLInteger":
        return 'int'
    elif xmltype == "XMLDouble":
        return 'float'
    elif xmltype == "XMLString":
        return 'str'
    elif xmltype == "XMLDate":
        return 'date'
    elif xmltype == "XMLDateTime":
        return 'datetime'
    else:
        return xmltype


def toinit(self, deep=False):
    """
        Nice init for complex types.

        All obligatory (non-nillable) children can also be created.

        Parameters
        ----------
        deep : bool, optional, defaule False
            If True all non-nillable children are created, otherwise
            they are simply None. The latter is used when
            converting response from XML to Python.
    """
    if not(deep):
        return
    for child in self._children:
        if child['min'] == 0:
            continue
        val_type = child['type']
        val = None
        if getattr(val_type, "_children", None) is not None:
            val = val_type(deep=deep)
        else:
            val = val_type()
        if child['max'].__class__.__name__ != "int" or child['max'] > 1:
            val = [val]
        setattr(self, child['name'], val)


def tostr(self):
    """
        Nice printing facility for complex types.
    """
    children = ''
    for child in self._children:
        child_name = child['name']
        array = ''
        if child['max'].__class__.__name__ != "int" or child['max'] > 1:
            array = '[]'
        child_value = getattr(self, child_name, None)
        many = False
        if len(array) and isinstance(child_value, (list, tuple)):
            many = True
        shift = len(child_name) + len(array) + 7  # 4 comes from tab
        if many:
            shift = shift + 1
            tmp = child_value
            stop = len(child_value)
            after = '\n]'
            if stop > 10:
                stop = 10
                after = '\n...' + after
            child_value = ''
            for val in tmp[:stop]:
                child_value = child_value + ',\n%s' % str(val)
            child_value = '[\n' + child_value[2:] + after
        elif child_value is not None:
            child_value = str(child_value)
        else:
            child_value = "%s (%s)" % (str(None),
                                       get_local_type(child['type'].__name__))
        child_value = child_value.replace('\n', '\n%s' % (' ' * shift))
        descr = '    %s%s = %s' % (child_name, array, child_value)
        children = children + '\n%s' % descr
    res = '(%s){%s\n}' % (self.__class__.__name__, children)

    return res


def equal(x1, x2):
    return x1.__dict__ == x2.__dict__


def notequal(x1, x2):
    return not(equal(x1, x2))


class XMLType(object):
    """
        Base xml schema type.

        It defines basic functions to_xml and from_xml.
    """
    _namespace = ""

    def to_xml(self, parent, name):
        """
            Function to convert to xml from python representation.

            This is basic function and it is suitable for complex types.
            Primitive types must overload it.

            Parameters
            ----------
            parent : etree.Element
                Parent xml element to append this child to.
            name : str
                Full qualified (with namespace) name of this element.
        """
        # this level element
        element = etree.SubElement(parent, name)

        # add all children to the current level
        # note that children include also base classes, as they are propagated by
        # the metaclass below
        for child in self._children:
            child_name = child["name"]
            full_child_name = child["fullname"]
            # get the value of the argument
            val = getattr(self, child_name, None)

            # do constraints checking
            n = 0  # number of values for constraints checking
            if hasattr(val, "__iter__") and val.__class__.__name__ != "str":
                n = len(val)
            elif val is not None:
                n = 1
                val = [val, ]

            if n < child["min"]:
                raise ValueError("Number of values for %s is less than "
                                 "min_occurs: %s" % (name, str(val)))
            if child["max"].__class__.__name__ == "int" and n > child["max"]:
                raise ValueError("Number of values for %s is more than max_occurs: %s" % (name, str(val)))

            if n == 0:
                continue  # only nillables can get so far

            # conversion
            for single in val:
                if not(hasattr(single, "to_xml")):
                    single = child['type'](single)
                single.to_xml(element, full_child_name)
                if child["type"] is XMLAny or \
                        (isinstance(single, XMLType) and type(single) != child["type"]):
                    # append type information
                    element[-1].set("{%s}type" % xmlnamespace.NS_XSI,
                                    etree.QName(
                                    "{%s}%s" % (single._namespace,
                                                single.__class__.__name__)))
                # try:
                    # single.to_xml(element, full_name)
                    # if child["type"] is XMLAny:
                        ## append type information
                        # element[-1].set("{%s}type" %xmlnamespace.NS_XSI,
                                #"{%s}%s" % (single._namespace,
                                           # single.__class__.__name__) )
                # except Exception:
                    # single = child['type'](single)
                    # single.to_xml(element, full_name)

    def from_xml(self, element):
        """
            Function to convert from xml to python representation.

            This is basic function and it is suitable for complex types.
            Primitive types must overload it.

            Parameters
            ----------
            element : etree.Element
                Element to recover from.
        """
        # removed with bug 7, we do not check for this for primitives,
        # so stay consistent for complex as well
        # element is nill
        #if element.get('{%s}nil' % xmlnamespace.NS_XSI, "false") == "true":
            #return

        all_children_names = []
        for child in self._children:
            all_children_names.append(child["name"])

        for subel in element:
            # name = xmlnamespace.get_local_name(subel.tag)
            name = subel.tag
            name = name[name.find("}")+1:]
            ind = all_children_names.index(name)

            # used for conversion. for primitive types we receive back built-ins
            inst = self._children[ind]['type']()
            explicit_type = subel.get('{%s}type' % xmlnamespace.NS_XSI)
            if explicit_type is not None:
                inst = XMLAny()

            # we do not distinguish xs:nil="true" explicitly here, this will have
            # empty text in any case, this is not strict standard, but ...
            subvalue = inst.from_xml(subel)

            # removed for bug 7
            #if subvalue is None:
                #if self._children[ind]['min'] != 0 and \
                   #self._children[ind]['nillable'] is False:
                    #raise ValueError("Non-nillable %s element is nil." % name)
            # None, i.e. nillables, should also be placed here 
            if self._children[ind]['max'].__class__.__name__ != "int" or\
               self._children[ind]['max'] > 1:
                current_value = getattr(self, name, None)
                if current_value is None:
                    current_value = []
                    setattr(self, name, current_value)
                current_value.append(subvalue)
            else:
                setattr(self, name, subvalue)
            del name, ind, inst

        # now all children were processed, so remove them to save memory
        element.clear()

        # do a simplistic validation that all expected elements were present,
        # this is not strict, but ...
        for child in self._children:
            val = getattr(self, child['name'], None)
            numValues = 0
            if val is not None:
                numValues = 1
            if isinstance(val, list):
                numValues = len(val)
            if numValues < child['min']:
                raise ValueError("Number of elements '%s' %d is less then minOccurs %d."\
                                 %(child['name'], numValues, child['min']))
            if child['max'].__class__.__name__ == "int" and\
               numValues > child['max']:
                raise ValueError("Number of elements '%s' %d is more then maxOccurs %d."\
                                 %(child['name'], numValues, child['max']))

        return self

    def to_file(self, fname):
        """
            Save to file as an xml string.

            Parameters
            ----------
            fname : str
                Filename to use.
        """
        if self._namespace:
            fullname = "{%s}%s" % (self._namespace, self.__class__.__name__)
        else:
            fullname = self.__class__.__name__
        root = etree.Element("root")
        self.to_xml(root, fullname)
        f = open(fname, "w")
        f.write(etree.tostring(root[0]).decode())
        f.close()

    @classmethod
    def from_file(cls, fname):
        """
            Create an instance from file.

            Parameters
            ----------
            fname : str
                Filename to parse.

            Returns
            -------
            out : new instance
        """
        f = open(fname)
        #s = f.read()
        #f.close()
        #root = etree.fromstring(s)
        root = xmlparser.parse_qualified(f)
        inst = cls()
        return inst.from_xml(root)


class ComplexTypeMeta(type):
    """
        Metaclass to create complex types on the fly.
    """
    def __new__(cls, name, bases, attributes):
        """
            Method to create new types.

            _children attribute must be present in attributes. It describes
            the arguments to be present in the new type. The he
            _children argument must be a list of the form:
            [{'name':'arg1', 'min':1, 'max':1, 'type':ClassType, "fullname":"name with ns"}, ...]
            Here fullname is used for serialization and must be qualified properly.

            Parameters
            ----------
            cls : this class
            name : str
                Name of the new type.
            bases : tuple
                List of bases classes.
            attributes : dict
                Attributes of the new type.
        """
        # list of children, even if empty, must be always present
        if not "_children" in attributes:
            attributes["_children"] = []

        # create dictionary for initializing class arguments
        clsDict = {}
        # iterate over children and add arguments to the dictionary
        # all arguments are initially have None value
        for attr in attributes["_children"]:
            # set the argument
            clsDict[attr['name']] = None
        # propagate documentation
        clsDict["__doc__"] = attributes.get("__doc__", None)
        # add nice printing
        clsDict["__str__"] = tostr
        clsDict["__repr__"] = tostr
        # add complex init
        clsDict["__init__"] = toinit
        # comparison
        clsDict["__eq__"] = equal
        clsDict["__ne__"] = notequal

        # extend children list with that of base classes
        new = []
        for b in bases:
            base_children = getattr(b, "_children", None)
            if base_children is not None:
                # append
                new.extend(base_children)
        new.extend(attributes["_children"])
        attributes["_children"] = new

        # children property is passed through
        clsDict["_children"] = attributes["_children"]

        # add ComplexType to base list
        if XMLType not in bases:
            newBases = list(bases)
            newBases.append(XMLType)
            bases = tuple(newBases)

        # propagate other non-reserved atributes
        for k in attributes:
            if k not in ("_children", "__init__", "__doc__",
                         "__ne__", "__eq__", "__str__", "__repr__"):
                clsDict[k] = attributes[k]

        # create new type
        return type.__new__(cls, name, bases, clsDict)

# the following is a modified copy from soaplib library


class XMLString(XMLType, str):

    def to_xml(self, parent, name):
        element = etree.SubElement(parent, name)
        element.text = unicode(self)

    def from_xml(self, element):
        if element.text:
            return element.text
        else:
            return ""


class XMLBase64Binary(XMLType, str):

    def to_xml(self, parent, name):
        element = etree.SubElement(parent, name)
        element.text = base64.b64encode(self)

    def from_xml(self, element):
        if element.text:
            return base64.b64decode(element.text)
        else:
            return ""


class XMLInteger(XMLType, int):
    def to_xml(self, parent, name):
        element = etree.SubElement(parent, name)
        element.text = repr(self)

    def from_xml(self, element):
        if element.text:
            try:
                return int(element.text.strip("L"))  # py2 strings can have L
            except:
                return long(element.text)
        return 0


class XMLDouble(XMLType, float):

    def to_xml(self, parent, name):
        element = etree.SubElement(parent, name)
        element.text = repr(self)

    def from_xml(self, element):
        if element.text:
            return float(element.text)
        return 0


class XMLBoolean(XMLType, str):

    def to_xml(self, parent, name):
        element = etree.SubElement(parent, name)
        if self in ('True', 'true', '1'):
            element.text = repr(True).lower()
        else:
            element.text = repr(False).lower()

    def from_xml(cls, element):
        if element.text:
            return (element.text.lower() in ['true', '1'])
        return False


class XMLAny(XMLType, str):

    _types = {}  # dict of known types

    def to_xml(self, parent, name):
        value = etree.fromstring(self)
        element = etree.SubElement(parent, name)
        element.append(value)

    def from_xml(self, element):
        # try to find types
        type = element.get('{%s}type' % xmlnamespace.NS_XSI, None)
        if type is None:
            return element
        type_class = self._types.get(type, None)
        if type_class is not None:
            res = type_class()
            return res.from_xml(element)
        else:
            return element


class XMLDecimal(XMLType, Decimal):

    def to_xml(self, parent, name):
        element = etree.SubElement(parent, name)
        element.text = str(self)

    def from_xml(self, element):
        if element.text:
            return Decimal(element.text)
        return Decimal(0)


class XMLDate(XMLType):

    def __init__(self, *arg):
        if len(arg) == 1 and isinstance(arg[0], date):
            self.value = arg[0]
        else:
            self.value = date(2008, 11, 11)

    def to_xml(self, parent, name):
        element = etree.SubElement(parent, name)
        element.text = self.value.isoformat()

    def from_xml(self, element):
        """expect ISO formatted dates"""
        if not(element.text):
            return date(1970, 1, 1)
        text = element.text
        y, m, d = text.split("-")[:3]
        y = int(y)
        m = int(m)
        d = int(d[:2])
        # ignore time zone information here
        return  date(y, m, d)


class XMLDateTime(XMLType):

    def __init__(self, *arg):
        if len(arg) == 1 and isinstance(arg[0], datetime):
            self.value = arg[0]
        else:
            self.value = datetime(2008, 11, 11)

    def to_xml(self, parent, name):
        element = etree.SubElement(parent, name)
        element.text = self.value.isoformat('T')

    def from_xml(self, element):
        if not(element.text):
            return datetime(1970, 1, 1)
        text = element.text
        # this way looks a bit slow, please complain if you need
        datestr, timestr = text.split("T",1)
        year, month, day = datestr.split("-")
        year = int(year)
        month = int(month)
        day = int(day)
        hour, minute, second = timestr.split(":", 2)
        hour = int(hour)
        minute = int(minute)
        rest = second[2:]
        second = int(second[:2])
        fraction = 0 
        if rest and rest[0] == ".":
            # fraction of second
            pos = len(rest)
            for i in range(1,len(rest)):
                if not rest[i].isdigit():
                    pos = i
                    break
            fraction = int(float(rest[:pos])*1e6)
            rest = rest[pos:]
        value = datetime(year, month, day, hour, minute, second, fraction)
        # time zone to UTC
        if rest and (rest[0] == "+" or rest[0] == "-"):
            zh, zm = rest.split(":",1)
            zh = int(zh)
            zm = int(rest[0]+zm[:2]) # add sign to minutes
            delta = timedelta(hours = zh, minutes = zm)
            value = value - delta
        return value

class XMLStringEnumeration(XMLType):
    _allowedValues = []

    def __init__(self, *arg):
        if len(arg) == 0:
            self.value = ""
        else:
            self.value = str(arg[0])

    def to_xml(self, parent, name):
        # putting this check here is a hack, to allow the complex type conversion to work properly here, since
        # it creates an instance
        if self.value not in self._allowedValues:
            raise ValueError("Not allowed value for this enumeration: value = %s" % (self.value))
        element = etree.SubElement(parent, name)
        element.text = unicode(self.value)

    def from_xml(self, element):
        val = ""
        if element.text:
            val = element.text
        if val not in self._allowedValues:
            raise ValueError("Not allowed value for this enumeration: value = %s" % (val))
        return val

# a map of primitive types
primmap = {
    'anyType':                                  XMLAny,
    '{%s}anyType' % xmlnamespace.NS_XSD:        XMLAny,
    'boolean':                                  XMLBoolean,
    '{%s}boolean' % xmlnamespace.NS_XSD:        XMLBoolean,
    'decimal':                                  XMLDecimal,
    '{%s}decimal' % xmlnamespace.NS_XSD:        XMLDecimal,
    'int':                                      XMLInteger,
    '{%s}int' % xmlnamespace.NS_XSD:            XMLInteger,
    'integer':                                  XMLInteger,
    '{%s}integer' % xmlnamespace.NS_XSD:        XMLInteger,
    'positiveInteger':                          XMLInteger,
    '{%s}positiveInteger' % xmlnamespace.NS_XSD: XMLInteger,
    'unsignedInt':                              XMLInteger,
    '{%s}unsignedInt' % xmlnamespace.NS_XSD:    XMLInteger,
    'nonNegativeInteger':                       XMLInteger,
    '{%s}nonNegativeInteger' % xmlnamespace.NS_XSD: XMLInteger,
    'short':                                    XMLInteger,
    '{%s}short' % xmlnamespace.NS_XSD:          XMLInteger,
    'byte':                                     XMLInteger,
    '{%s}byte' % xmlnamespace.NS_XSD:           XMLInteger,
    'unsignedByte':                             XMLInteger,
    '{%s}unsignedByte' % xmlnamespace.NS_XSD:   XMLInteger,
    'long':                                     XMLInteger,
    '{%s}long' % xmlnamespace.NS_XSD:           XMLInteger,
    'unsignedLong':                             XMLInteger,
    '{%s}unsignedLong' % xmlnamespace.NS_XSD:   XMLInteger,
    'float':                                    XMLDouble,
    '{%s}float' % xmlnamespace.NS_XSD:          XMLDouble,
    'double':                                   XMLDouble,
    '{%s}double' % xmlnamespace.NS_XSD:         XMLDouble,
    'string':                                   XMLString,
    '{%s}string' % xmlnamespace.NS_XSD:         XMLString,
    'base64Binary':                             XMLBase64Binary,
    '{%s}base64Binary' % xmlnamespace.NS_XSD:   XMLBase64Binary,
    'anyURI':                                   XMLString,
    '{%s}anyURI' % xmlnamespace.NS_XSD:         XMLString,
    'language':                                 XMLString,
    '{%s}language' % xmlnamespace.NS_XSD:       XMLString,
    'token':                                    XMLString,
    '{%s}token' % xmlnamespace.NS_XSD:          XMLString,
    'date':                                     XMLDate,
    '{%s}date' % xmlnamespace.NS_XSD:           XMLDate,
    'dateTime':                                 XMLDateTime,
    '{%s}dateTime' % xmlnamespace.NS_XSD:       XMLDateTime,
    # FIXME: probably timedelta, but needs parsing.
    # It looks like P29DT23H54M58S
    'duration':                                 XMLString,
    '{%s}duration' % xmlnamespace.NS_XSD:        XMLString}
XMLAny._types = primmap.copy()



# xmlschema.py - XMLSchemaParser class, part of osa.
# Copyright 2013 Sergey Bozhenkov, boz at ipp.mpg.de
# Licensed under LGPLv3 or later, see the COPYING file.

'''
    Conversion of XML Schema types into Python classes.
'''
from . import xmlnamespace
from . import xmltypes
from . import xmlparser


class XMLSchemaParser(object):
    '''
        Parser to get types from an XML Schema.
    '''
    def __init__(self, root, wsdl_url=None):
        '''
            Initialize parser.

            self.schema - the root node of the schema
            self.tns - target namespace
            self.imported - a list of parsers for imported schemas
            self.qualified - value of elementFormDefault, used
                             for handling namespaces

            Parameters
            ----------
            root : xml schema, i.e. <schema ...> ... </schema>
                The schema to parse
            wsdl_url : string
                Url the wsdl comes from, can be used to expand schema
                references.
        '''
        if wsdl_url is None:
            wsdl_url = ''
        self.wsdl_url = wsdl_url
        # check we have a schema
        if root.tag != '{%s}schema' % xmlnamespace.NS_XSD:
            raise ValueError('Supplied root node \'%s\' is not of XML Schema '
                             'type.' % root.tag)

        # set schema parameters
        self.schema = root
        self.tns = self.schema.get('targetNamespace', '')
        self.qualified = self.schema.get('elementFormDefault', 0)
        if self.qualified == 'qualified':
            self.qualified = 1

        # find and initialize imported ones
        self.imported = []
        imports = self.schema.findall('.//{%s}import' % xmlnamespace.NS_XSD)
        imports.extend(self.schema.findall('.//{%s}include' %
                                           xmlnamespace.NS_XSD))
        for schema in imports:
            loc = schema.get('schemaLocation', None)
            # basically says that types from that namespace will be used, no real
            # import, i.e. the real schema was defined already
            if loc is None:
                continue
            # standard namespace we know by default
            if loc in (xmlnamespace.NS_SOAP, xmlnamespace.NS_SOAP12,
                       xmlnamespace.NS_SOAP_ENC, xmlnamespace.NS_SOAP_ENV,
                       xmlnamespace.NS_WSDL, xmlnamespace.NS_XSD,
                       xmlnamespace.NS_XSI):
                continue
            # try getting the schema
            parser = XMLSchemaParser(xmlparser.parse_qualified_from_url(loc, wsdl_url=wsdl_url), wsdl_url=wsdl_url)
            # check if want to change the schema namespace
            ns = schema.get('namespace', None)
            if ns is not None:
                parser.tns = ns
            self.imported.append(parser)

    def generate_classes(self):
        '''
            Generate Python classes from this schema.

            Returns
            -------
            out : dictionary
                Dictionary of types {ns}name -> Python class
        '''
        xlist = self.get_list_of_defined_types()
        types = XMLSchemaParser.convert_xmltypes_to_python(xlist)
        return types

    def get_list_of_defined_types(self):
        '''
            Construct a dictionary: type name -> xml node

            Types are given by complexType, simpleType or element.
            Types from imported schemas are included as well.
            Type names include namespaces.

            Returns
            -------
            out : dict
                A dictionary of defined types.
        '''
        # get list of types
        raw = self.schema.findall('./{%s}complexType' % xmlnamespace.NS_XSD)
        raw.extend(self.schema.findall('./{%s}simpleType' % xmlnamespace.NS_XSD))
        #<element> entries are usually used in message section in
        # wsdl. Such an entry can either define its type inside,
        # or simply be an alias to defined type. However. we
        # can not simply attach all elements to types, because
        # element names can shadow the types. Therefore, do it
        # one by one.
        elements = self.schema.findall('./{%s}element' % xmlnamespace.NS_XSD)

        # create dictionary by getting also names of the types
        types = {}
        for el in raw:
            name = el.get('name', None)
            if name is not None:  # consider an exception
                name = '{%s}%s' % (self.tns, name)
                el.set('qualified', self.qualified)
                types[name] = el

        for el in elements:
            name = el.get('name', None)
            if name is not None:
                name = '{%s}%s' % (self.tns, name)
                el.set('qualified', self.qualified)
                if not name in types:
                    types[name] = el

        # go over all children and append their types
        for parser in self.imported:
            types.update(parser.get_list_of_defined_types())

        return types

    @staticmethod
    def convert_xmltypes_to_python(xtypes):
        '''
            Convert xml types definitions in the dictionary
            into Python classes.

            Parameters
            ----------
            xtypes : dictionary name -> xml element
                A dictionary as returned by get_list_of_defined_types.

            Returns
            -------
            out : dictionary name -> Python class
        '''
        types = {}
        for k in xtypes:
            # if the class was already created as a parent of another class, do nothing
            if k in types:
                continue
            x = xtypes[k]
            XMLSchemaParser.create_type(k, x, xtypes, types)
        return types

    @staticmethod
    def create_type(name, element, xtypes, types):
        '''
            Creates proper type for the element.

            The created types is appended to the types.

            Parameters
            ----------
            name : str
                Class name
            element : xml element
                Class node.
            xtypes : dictionary class name -> xml node
            types : dictionary class name -> Python class
                The result is appended here.
        '''
        # I need this a a separate function to be able to call
        # it recursively for parents and children
        # a switch to decide what to do
        # element.complexType = complexType
        # element.simpleType = simpleType
        # element empty with type = xx = an alias to xx
        # element empty - empty class
        # complexType.sequence/all/choice - complex class, no parent
        # complexContent.extension - complex class with parent
        # simpleType.restriction - e.g. string enumeration
        if element.tag == '{%s}element' % xmlnamespace.NS_XSD:
            children = element.findall('./{%s}complexType' % xmlnamespace.NS_XSD)
            children.extend(element.findall('./{%s}simpleType' % xmlnamespace.NS_XSD))
            if len(children) == 1:
                children[0].set('qualified', element.get('qualified', 0))
                element = children[0]
            elif len(children) == 0:
                type = element.get('type', None)
                if type is not None:
                    # alias
                    XMLSchemaParser.create_alias(name, type, xtypes, types)
                    return
                else:
                    # empty class
                    XMLSchemaParser.create_empty_class(name, types)
                    return
            else:
                raise ValueError(' Wrong schema structure. Element has more than 1 '
                        ' child specifying the class. Element: %s' % (name))
        if element.tag == '{%s}complexType' % xmlnamespace.NS_XSD:
            # complex class
            XMLSchemaParser.create_complex_class(name, element, xtypes, types)
        elif element.tag == '{%s}simpleType' % xmlnamespace.NS_XSD and\
                element[0].tag == '{%s}restriction' % xmlnamespace.NS_XSD \
                and element[0].get('base', None) == '{%s}string' % xmlnamespace.NS_XSD \
                and len(element[0]) > 0 \
                and element[0][0].tag == '{%s}enumeration' % xmlnamespace.NS_XSD:
            XMLSchemaParser.create_string_enumeration(name, element, types)
        elif element.tag == '{%s}simpleType' % xmlnamespace.NS_XSD and\
                element[0].tag == '{%s}restriction' % xmlnamespace.NS_XSD \
                and element[0].get('base', None) is not None:
            base_type = element[0].get('base')
            XMLSchemaParser.create_alias(name, base_type, xtypes, types)

    @staticmethod
    def get_type_by_name(name, xtypes, types):
        '''
            Return requested class from primmap or as created from xml.

            Parameters
            ----------
            name : str
                Type name.
            xtypes : dict
                List of xml elements to look in.
            types : dict
                List of already created classes to look in.

            Returns
            -------
            out : class
        '''
        if name is None:
            return
        elif name in xmltypes.primmap:
            cl = xmltypes.primmap[name]
        elif name in types:
            cl = types[name]
        elif not name in xtypes:
            raise ValueError(' Class %s not found in anywhere' % (name))
        else:
            XMLSchemaParser.create_type(name, xtypes[name], xtypes, types)
            cl = types[name]
        return cl

    @staticmethod
    def get_doc(x):
        '''
            Extract documentation from element.

            Parameters
            -----------
            x : xml element

            Returns
            -------
            out : str
                Documentation from whatever found <documentation> out </documentation>
        '''
        doc = x.find('.//{%s}documentation' % xmlnamespace.NS_XSD)
        if doc is None:
            doc = x.find('.//documentation')
        if doc is not None:
            return doc.text
        else:
            return 'no documentation'

    @staticmethod
    def create_alias(name, alias_type, xtypes, types):
        '''
            Create a copy of known class with proper namespace.

            Parameters
            ----------
            name : str
                Name of the new class.
            alias_type : str
                The target alias
            xtypes : dictionary class name -> xml node
            types : dictionary of classes
                The new aliases is appended here.
        '''
        alias = XMLSchemaParser.get_type_by_name(alias_type, xtypes, types)
        if alias is None:
            return
        cls_name = xmlnamespace.get_local_name(name)
        cls_ns = xmlnamespace.get_ns(name)
        # create new type since the namespace may be different
        cls = type(cls_name, (alias,), {'__doc__': 'no documentation',
                                        '_namespace': cls_ns})
        types[name] = cls

    @staticmethod
    def create_empty_class(name, types):
        '''
            Create empty class, i.e. no children.

            Parameters
            ----------
            name : str
                Name of the new class.
            alias_type : str
                The target alias
            xtypes : dictionary class name -> xml node
            types : dictionary of classes
                The new aliases is appended here.
        '''
        cls_name = xmlnamespace.get_local_name(name)
        cls_ns = xmlnamespace.get_ns(name)
        cls = xmltypes.ComplexTypeMeta(cls_name, [],
                                       {'_children': [],
                                        '__doc__': 'no documentation',
                                        '_namespace': cls_ns})
        types[name] = cls

    @staticmethod
    def create_string_enumeration(name, element, types):
        '''
            Creates a copy of XMLStringEnumertion with properly set
            allowed values.

            The created class is attached to types.

            Parameters
            ----------
            name : str
                Name of the new class.
            element : `etree.Element`
                XML description of the enumeration
            types : dictionary of classes
        '''
        xvalues = element.findall('.//{%s}enumeration' % xmlnamespace.NS_XSD)
        values = []
        for x in xvalues:
            values.append(x.get('value', None))
        doc = XMLSchemaParser.get_doc(element)
        # create new class
        # I choose to give short names to classes, i.e. without
        # a namespace, even though Python can manage full names as well
        cls_name = xmlnamespace.get_local_name(name)
        cls_ns = xmlnamespace.get_ns(name)
        cls = type(cls_name, (xmltypes.XMLStringEnumeration,),
                   {'_allowedValues': values, '__doc__': doc,
                    '_namespace': cls_ns})
        types[name] = cls

    @staticmethod
    def create_complex_class(name, element, xtypes, types):
        '''
            Create complex class.

            Parameters
            ----------
            name : str
                Class name
            element : xml element
                Class node.
            xtypes : dictionary class name -> xml node
            types : dictionary class name -> Python class
                The result is appended here.
        '''
        qualified = element.get('qualified', 0)
        # decide if we have a parent and first create that
        parents = []
        exts = element.findall('./{%s}complexContent/{%s}extension' % (xmlnamespace.NS_XSD, xmlnamespace.NS_XSD))
        if exts is not None:
            for ext in exts:
                parent_name = ext.get('base', None)
                parent = XMLSchemaParser.get_type_by_name(parent_name, xtypes, types)
                parents.append(parent)

        # find sequence/choice/all
        seq = None
        for str in ('sequence', 'all', 'choice'):
            # note deep search here, this is necessary for
            # extensions that look like
            # complexType->complexContent->extensions->sequence
            seq = element.find('.//{%s}%s' % (xmlnamespace.NS_XSD, str))
            if seq is not None:
                break

        # collect children
        cls_ns = xmlnamespace.get_ns(name)
        children = []
        if seq is not None:
            for s in seq:
                # iterate over sequence, do not consider in place defs
                ref = s.get('ref', None)  # reference to another element
                if ref is not None:
                    # can there be a reference to a local element?
                    # I assume not => always qualified
                    type_name = ref
                    # child_name = xmlnamespace.get_local_name(ref)
                    child_name = ref  # name is qualified by the target, is it correct?
                else:
                    type_name = s.get('type', None)
                    child_name = s.get('name', 'unknown')

                if type_name is None:
                    compl = s.find('./{%s}complexType' % (xmlnamespace.NS_XSD))
                    simpl = s.find('./{%s}simpleType' % (xmlnamespace.NS_XSD))
                    if compl is not None:
                        XMLSchemaParser.create_type(child_name, compl, xtypes, types)
                        type_name = child_name
                    elif simpl is not None:
                        XMLSchemaParser.create_type(child_name, simpl, xtypes, types)
                        type_name = child_name
                    else:
                        continue
                type = XMLSchemaParser.get_type_by_name(type_name, xtypes, types)
                minOccurs = int(s.get('minOccurs', 1))
                maxOccurs = s.get('maxOccurs', 1)
                if maxOccurs != 'unbounded':
                    maxOccurs = int(maxOccurs)
                nillable = s.get('nillable', False) == 'true'
                # child name is qualified as required, i.e. it is a full name
                full_child_name = child_name
                # class member names!
                child_name = xmlnamespace.get_local_name(child_name)
                if qualified and full_child_name.find('}') == -1:
                    full_child_name = '{%s}%s' % (cls_ns, full_child_name)
                children.append({'name': child_name, 'type': type,
                                 'min': minOccurs, 'max': maxOccurs,
                                 'nillable': nillable,
                                 'fullname': full_child_name})
        # get doc
        doc = XMLSchemaParser.get_doc(element)

        # create new class
        # I choose to give short names to classes, i.e. without
        # a namespace, even though Python can manage full names as well
        cls_name = xmlnamespace.get_local_name(name)
        cls = xmltypes.ComplexTypeMeta(cls_name, parents,
                                       {'_children': children, '__doc__': doc,
                                        '_namespace': cls_ns})
        types[name] = cls



# message.py - Message class, part of osa.
# Copyright 2013 Sergey Bozhenkov, boz at ipp.mpg.de
# Licensed under LGPLv3 or later, see the COPYING file.

"""
   Python class for input/output messages.
"""
# from . import xmlnamespace
from . import xmltypes
# import xml.etree.cElementTree as etree


class Message(object):
    """
        Message for input and output of service operations.

        Messages perform conversion of Python to xml and backwards
        of the calls and returns.

        At the moment only document/literal wrapped is implemented.

        Parameters
        ----------
        name : str
            Namespace qualified name of the message.
        parts : list
            List of message parts in the form
            (part name, part type class).
            This description is usually found in message part of a WSDL document.
            Note, that due to binding section not all message parts are used for encoding.
            The parts that are used are given be use_parts.
        use_parts : list
            List of parts to be really used for encoding/decoding.
            This comes from wsdl binding section. Yes, they are not
            quite from this planet.
            In any case, in the present implementation I assume doc/literal wrapped and
            use only the very first part from this member for encoding.
    """
    def __init__(self, name, parts, use_parts=None):
        self.name = name
        self.parts = parts
        if use_parts is None:
            use_parts = []
        self.use_parts = use_parts

    def __str__(self, switch="wrap"):
        """
            String representation of the message in three forms:
                - wrapped message
                - positional sub-arguments
                - keyword sub-arguments.
                - out - the only child of wrapped message. This applicable
                        to output message extraction.

            Parameters
            ----------
            switch : str, optional
                Specifies which form to return: wrap, positional, keyword, out.
        """
        # assumed wrapped convention
        if len(self.use_parts) < 1:
            return ""
        p = self.use_parts[0][1]  # message type
        res = ''
        if switch == "positional":
            for child in getattr(p, "_children", []):
                opt = ''
                array = ''
                if child['max'].__class__.__name__ != "int" or child['max'] > 1:
                    array = '[]'
                if child['min'] == 0:
                    opt = '| None'
                type = xmltypes.get_local_type(child['type'].__name__)
                res = '%s, %s%s %s %s' % (res, type, array, child["name"], opt)
        elif switch == "keyword":
            for child in getattr(p, "_children", []):
                opt = ''
                array = ''
                if child['max'].__class__.__name__ != "int" or child['max'] > 1:
                    array = '[]'
                if child['min'] == 0:
                    opt = '| None'
                type = xmltypes.get_local_type(child['type'].__name__)
                res = '%s, %s=%s%s %s' % (res, child['name'], type, array, opt)
        elif switch == 'out' and len(getattr(p, "_children", [])) == 1:
            child = p._children[0]
            opt = ''
            array = ''
            if child['max'].__class__.__name__ != "int" or child['max'] > 1:
                array = '[]'
            if child['min'] == 0:
                opt = '| None'
            type = xmltypes.get_local_type(child['type'].__name__)
            res = '%s%s %s %s' % (type, array, 'result', opt)
        else:
            res = '%s %s' % (p.__name__, 'msg')

        if len(res) > 2 and res[0] == ',':
            res = res[2:]

        return res

    def to_xml(self, *arg, **kw):
        """
            Convert from Python into xml message.

            This function accepts parameters as they are supplied
            to the method call and tries to convert it to a message.
            Arguments can be in one of  four forms:
                - 1 argument of proper message type for this operation
                - positional arguments - members of the proper message type
                - keyword arguments - members of the message type.
                - a mixture of positional and keyword arguments.

            Keyword arguments must have at least one member: _body which
            contains etree.Element to append the conversion result to.
        """
        if len(self.use_parts) < 1:
            # etree.SubElement(kw["_body"], self.name)
            return
        # assumed wrapped convention
        cl = self.use_parts[0][1]  # class
        p = cl()  # encoding instance

        # wrapped message is supplied
        if len(arg) == 1 and isinstance(arg[0], cl):
            for child in getattr(p, "_children", []):
                p = arg[0]
        else:
            # reconstruct wrapper from expanded input
            counter = 0
            for child in getattr(p, "_children", []):
                name = child["name"]
                # first try keyword
                val = kw.get(name, None)
                if val is None:  # not keyword
                    if counter < len(arg):
                        # assume this is positional argument
                        val = arg[counter]
                        counter += 1
                if val is None:  # check if nillable
                    if child["min"] == 0:
                        continue
                    else:
                        raise ValueError("Non-nillable parameter %s is not "
                                         "present" % name)
                setattr(p, name, val)

        # set default ns to save space
        # this does not work with xml qualified/unqualified, need a hack
        # etree.register_namespace("", xmlnamespace.get_ns(self.name))
        # the real conversion is done by ComplexType
        # messages always refer to a top level element => qualified
        p.to_xml(kw["_body"], "{%s}%s" % (p._namespace, p.__class__.__name__))

    def from_xml(self, body, header=None):
        """
            Convert from xml message to Python.
        """
        if len(self.use_parts) < 1:
            return None
        # assumed wrapped convention
        p = self.use_parts[0][1]()  # decoding instance

        res = p.from_xml(body)

        # for wrapped doc style (the only one implemented) we know, that
        # wrapper has only one child, get it
        if len(getattr(p, "_children", [])) == 1:
            return getattr(res, p._children[0]["name"], None)
        else:
            return res



# method.py - Method class, part of osa.
# Copyright 2013 Sergey Bozhenkov, boz at ipp.mpg.de
# Licensed under LGPLv3 or later, see the COPYING file.

"""
    SOAP operation class.
"""
from . import xmlnamespace
from . import xmlparser
import sys
if sys.version_info[0] < 3:
    from urllib2 import urlopen, Request, HTTPError
else:
    from urllib.request import urlopen, Request, HTTPError
import xml.etree.cElementTree as etree

# some standard stuff
SOAP_BODY = '{%s}Body' % xmlnamespace.NS_SOAP_ENV
SOAP_FAULT = '{%s}Fault' % xmlnamespace.NS_SOAP_ENV
SOAP_HEADER = '{%s}Header' % xmlnamespace.NS_SOAP_ENV


class Method(object):
    """
        Definition of a single SOAP method, including location, action, name
        and input and output classes.

        Parameters
        ----------
        name : str
            Name of operation
        input : `osa.message.Message` instance
            Input message.
        output : `osa.message.Message` instance
            Output message.
        doc : str, optional - default to None
            Documentation of the method as found in portType section of WSDL.
        action : str
            Soap action string.
        location : str
            Location as found in service part of WSDL.
    """
    def __init__(self, name, input, output, doc=None,
                 action=None, location=None):
        self.name = name
        self.input = input
        self.output = output
        self.location = location
        self.action = action
        self._doc = doc
        self._redoc()

    def _redoc(self):
        """
            Add call signatures to doc.
        """
        sign = '%s\n%s\n%s' % (self.__str__(),
                               self.__str__(switch="positional"),
                               self.__str__(switch="keyword"))
        self.__doc__ = '%s\n%s' % (sign, self._doc)

    def __str__(self, switch='wrap'):
        """
            String representation of the call in three forms:
                - wrapped message
                - positional sub-arguments
                - keyword sub-arguments.

            Parameters
            ----------
            switch : str, optional
                Specifies which form to return: wrap, positional, keyword.
        """
        input_msg = self.input.__str__(switch=switch)
        if self.output is None:
            output_msg = "None"
        else:
            output_msg = self.output.__str__(switch='out')

        return '%s = %s(%s)' % (output_msg, self.name, input_msg)

    def __call__(self, *arg, **kw):
        """
            Process rpc-call.
        """
        # create soap-wrap around our message
        env = etree.Element('{%s}Envelope' % xmlnamespace.NS_SOAP_ENV)
        # header = etree.SubElement(env, '{%s}Header' % xmlnamespace.NS_SOAP_ENV)
        body = etree.SubElement(env, '{%s}Body' % xmlnamespace.NS_SOAP_ENV)

        # compose call message - convert all parameters and encode the call
        kw["_body"] = body
        self.input.to_xml(*arg, **kw)

        text_msg = etree.tostring(env)  # message to send
        del env

        # http stuff
        request = Request(self.location, text_msg,
                          {'Content-Type': 'text/xml',
                           'SOAPAction': self.action})
        del text_msg

        # real rpc
        try:
            response = urlopen(request)
            del request
            # check http code returned
            if response.code == 200:
                if self.output is None:
                    return None
                # string to xml
                # use qualified parsing to get the anyType
                # type properly, unless it hits the performance heavily
                xml = xmlparser.parse_qualified(response)
                # response = response.read()
                # xml = etree.fromstring(response)
                del response
                # find soap body
                body = xml.find(SOAP_BODY)
                if body is None:
                    raise RuntimeError("No SOAP body found in response")
                body = body[0]
                return self.output.from_xml(body)
            elif response.code == 202 or response.code == 204 \
                    and self.output is None:
                return None
            else:
                raise RuntimeError("Bad HTTP status code: %d" % response.code)
        except HTTPError as e:
            if e.code == 500:
                # read http error body and make xml from it
                try:
                    xml = etree.fromstring(e.fp.read())
                except Exception:
                    print(e)
                    return None
                body = xml.find(SOAP_BODY)
                if body is None:
                    raise
                # process service fault
                fault = body.find(SOAP_FAULT)
                if fault is not None:
                    code = fault.find('faultcode')
                    if code is not None:
                        code = code.text or ''
                    string = fault.find('faultstring')
                    if string is not None:
                        string = string.text or ''
                    detail = fault.find('detail')
                    if detail is not None:
                        detail = detail.text or ''
                    raise RuntimeError("SOAP Fault %s: %s <%s> %s %s" %
                                       (self.location, self.name, code,
                                        string, detail))
                else:
                    raise
            else:
                raise



# client.py - client class, part of osa.
# Copyright 2013 Sergey Bozhenkov, boz at ipp.mpg.de
# Licensed under LGPLv3 or later, see the COPYING file.

"""
    Top level access to SOAP service.
"""

from . import xmlnamespace
from . import wsdl


def str_for_containers(self):
    """
        Nice printing for types and method containers.

        Containers must have _container attribute containing all
        elements to be printed.
    """
    cont = getattr(self, '_container', None)
    if cont is None:
        return ''
    res = ''
    for child in cont:
        descr = str(getattr(getattr(self, child, None), '__doc__', None))
        if len(descr) > 100:
            descr = descr[:100] + '...'
        descr = descr.replace('\n', '\n\t')
        res = res + '\n%s\n\t%s' % (child, descr)
    res = res[1:]
    return res


class Client(object):
    """
        Top level class to talk to soap services.

        This is an access point to service functionality. The client accepts
        WSDL address and uses `osa.wsdl.WSDLParser` to get all defined types and
        operations. The types are set to client.types and operations
        are set to self.service.

        To examine present types or operations simply print (or touch repr)::

            >>> client.types

        or::

            >>> client.service

        correspondingly.

        To create type simply call::

            >>> client.types.MyTypeName().

        Class constructor will also create all obligatory (non-nillable) children.
        To call an operation::

            >>> client.service.MyOperationName(arg1, arg2, arg3, ...),

        where arguments are of required types. Arguments can also
        be passed as keywords or a ready wrapped message.

        If any help is available in the WSDL document it is propagated to the
        types and operations, see e.g. help
        :py:attr:`client.types.MyTypeName`. In addition
        the help page on an operation displays its call signature.

        Nice printing is also available for all types defined in client.types::

            >>> print(client.types.MyTypeName())

        .. warning::
            Only document/literal wrapped convention is implemented at the moment.

        In reality client.types and client.service are simply containers.
        The content of these containers is set from results of parsing
        the wsdl document by
        `osa.wsdl.WSDLParser.get_types` and
        `osa.wsdl.WSDLParser.get_services` correspondingly.
        See also `osa.wsdl.WSDLParser.parse`.

        The client.types container consists of auto generated (by
        `osa.xmlschema.XMLSchemaParser`)
        class definitions. So that a call to a member returns and instance
        of the new type. New types are auto-generated according to a special
        convention by metaclass `osa.xmltypes.ComplexTypeMeta`.

        The client.service container consists of methods wrapers
        methods.Method. The method wrapper is callable with free number of
        parameters. The input and output requirements of a method are
        contained in methods.Message instances `osa.methods.Method.input` and
        `osa.methods.Method.output` correspondingly. On a call a method converts
        the input to XML by using Method.input, sends request to the
        service and finally decodes the response from XML by
        Method.output.

        Parameters
        ----------
        wsdl_url : str
            Address of wsdl document to consume.
    """
    def __init__(self, wsdl_url):
        #create parser and download the WSDL document
        self.wsdl_url = wsdl_url
        parser = wsdl.WSDLParser(wsdl_url)
        self._types, self._services = parser.parse()
        self.names = []
        self.create_types_container()
        self.create_services_containers()

        return

    def create_types_container(self):
        """
            Create types container class for easy access.

            As a result of this method, self.types contains
            all the defined classes with their short names,
            i.e. without namespace prefix. If a name collision
            is detected, the second and all consecutive classes
            are appended with a counter.
        """
        types = {}
        for k, v in self._types.items():
            short_name = xmlnamespace.get_local_name(k)
            if short_name in types:
                counter = 1
                while True:
                    new_name = '%s_%d' % (short_name, counter)
                    counter += 1
                    if not new_name in types:
                        short_name = new_name
                        break
            types[short_name] = v
        types['_container'] = list(types)
        types['__str__'] = str_for_containers
        types['__repr__'] = str_for_containers
        self.types = type('TypesDispatcher', (), types)()

    def create_services_containers(self):
        """
            Create methods containers for easy access.

            As a result of this method, self.service
            with available operations is created. If
            there are several services in the supplied
            wsdl, than self.service_1, self.service_2
            are created.
        """
        def create(name, attr_name, methods):
            """
                Do create.

                Parameters
                ----------
                name : nice server name
                attr_name : hot to attach service, i.e. self.service_1
                methods : dict of service methods
            """
            methods['_container'] = list(methods)
            methods['__str__'] = str_for_containers
            methods['__repr__'] = str_for_containers
            setattr(self, attr_name,
                    type('ServiceDispatcher', (), methods)())
            self.names.append('%s %s' % (attr_name, name))

        if len(self._services) == 1:
            l = list(self._services)
            create(l[0], 'service', self._services[l[0]])
        else:
            counter = 1
            for k, v in self._services.items():
                attr_name = 'service_%d' % counter
                counter += 1
                create(k, attr_name, v)

    def __str__(self):
        res = ''
        for name in self.names:
            res = res + ', %s' % name
        res = res[2:] + ' from:\n\t%s' % (self.wsdl_url)
        return res

    def __repr__(self):
        return self.__str__()



# __init__.py - __init__ file, part of osa.
# Copyright 2013 Sergey Bozhenkov, boz at ipp.mpg.de
# Licensed under LGPLv3 or later, see the COPYING file.

__version__ = "0.2.3"

from .client import Client



# wsdl.py - WSDLParser class, part of osa.
# Copyright 2013 Sergey Bozhenkov, boz at ipp.mpg.de
# Licensed under LGPLv3 or later, see the COPYING file.

"""
    Conversion of WSDL documents into Python.
"""
from . import xmlnamespace
from . import xmlparser
from . import xmlschema
from . import xmltypes
from . import message
from . import method
import xml.etree.cElementTree as etree


class WSDLParser(object):
    """
        Parser to get types and methods defined in the document.
    """
    def __init__(self, wsdl_url):
        """
            Initialize parser.

            The WSDL document is loaded and is converted into xml.

            Initialized members:
            self.wsdl_url  - url of wsdl document, or file
            self.wsdl - xml document read from wsdl_url (etree.Element)
            self.tns - target namespace

            Parameters
            ----------
            wsdl_url : str
                Address of the WSDL document.
        """
        self.wsdl_url = wsdl_url
        self.wsdl = xmlparser.parse_qualified_from_url(wsdl_url)
        if self.wsdl.tag != "{%s}definitions" % xmlnamespace.NS_WSDL:
            raise ValueError("Not a WSDL xml, the top level element: %s" %
                             self.wsdl.tag)
        # get target namespace
        self.tns = self.wsdl.get('targetNamespace', "")

    def get_types(self, ):
        """
            Constructs a map of all types defined in the document.

            Returns
            -------
            out : dict
                A map of found types {type_name : complex class}
        """
        types_section = self.wsdl.findall('.//{%s}types' %
                                          xmlnamespace.NS_WSDL)[0]
        schemas = types_section.findall('./{%s}schema' % xmlnamespace.NS_XSD)
        xtypes = {}
        for schema in schemas:
            parser = xmlschema.XMLSchemaParser(schema, wsdl_url=self.wsdl_url)
            xtypes.update(parser.get_list_of_defined_types())
        types = xmlschema.XMLSchemaParser.convert_xmltypes_to_python(xtypes)
        xmltypes.XMLAny._types.update(types)
        return types

    def get_messages(self, types):
        """
            Construct messages from message section.

            Parameters
            ----------
            types : dictionary of types
                Types as returned by get_types().

           Returns
           -------
           out : dict
            Map message name -> Message instance
        """
        xmessages = self.wsdl.findall('./{%s}message' % xmlnamespace.NS_WSDL)
        messages = {}
        for x in xmessages:
            message_name = "{%s}%s" % (self.tns, x.get("name", ""))
            parts = []
            xparts = x.findall('./{%s}part' % xmlnamespace.NS_WSDL)
            for y in xparts:
                part_name = y.get("name", "")
                part_type = y.get("element", None)
                if part_type is None:
                    part_type = y.get("type", None)
                if part_type is None:
                    raise ValueError("Could not find part type in:\n %s"
                                     % (etree.tostring(x).decode()))
                cls = None
                if part_type in types:
                    cls = types[part_type]
                elif part_type in xmltypes.primmap:
                    cls = xmltypes.primmap[part_type]
                else:
                    raise ValueError("Type %s not found for message:\n%s" %
                                     (part_type, etree.tostring(x).decode()))
                parts.append([part_name, cls])
            messages[message_name] = message.Message(message_name,
                                                     parts)
        return messages

    def get_operations(self, messages):
        """
            Get list of operations with messages
            from portType section.

            Parameters
            ----------
            messages : dict
                Dictionary of message from `get_messages`.

           Returns
           -------
           out : dict
            {portType -> {operation name -> Method instance}}
            The method here does not have location.
        """
        xports = self.wsdl.findall('./{%s}portType' % xmlnamespace.NS_WSDL)
        ports = {}
        for xport in xports:
            port_name = "{%s}%s" % (self.tns, xport.get("name", ""))
            ports[port_name] = {}
            xops = xport.findall('./{%s}operation' % xmlnamespace.NS_WSDL)
            for xop in xops:
                op_name = xop.get("name", "")
                ports[port_name][op_name] = {}
                xin = xop.findall('./{%s}input' % xmlnamespace.NS_WSDL)
                if not(xin):
                    raise ValueError("No input message in operation: \n%s" %
                                     (etree.tostring(xop).decode()))
                in_name = xin[0].get("message", "")
                if not in_name in messages:
                    raise ValueError("Message %s not found." % in_name)
                in_cl = messages[in_name]
                out_cl = None
                xout = xop.findall('./{%s}output' % xmlnamespace.NS_WSDL)
                if xout:
                    out_name = xout[0].get("message", "")
                    if not out_name in messages:
                        raise ValueError("Message %s not found." % in_name)
                    out_cl = messages[out_name]

                # documentation
                doc = xop.find('{%s}documentation' % xmlnamespace.NS_WSDL)
                if doc is not None:
                    doc = doc.text

                op = method.Method(op_name, in_cl, out_cl, doc=doc)
                ports[port_name][op_name] = op
        return ports

    def get_bindings(self, operations):
        """
            Check binding document/literal and http transport.

            If any of the conditions is not satisfied
            the binding is dropped, i.e. not present in
            the return value. This also sets soapAction
            and use_parts of the messages.

            Parameters
            ----------
            operations : dict as returned by get_operations

            Returns
            -------
            out : dict
             Map similar to that from get_operations but
             with binding names instead of portType names.
        """
        xbindings = self.wsdl.findall("./{%s}binding" % xmlnamespace.NS_WSDL)
        bindings = {}
        for xb in xbindings:
            b_name = "{%s}%s" % (self.tns, xb.get("name", ""))
            b_type = xb.get("type", None)
            if b_type is None:
                raise ValueError("No type in binding %s" %
                                 (etree.tostring(xb).decode()))
            if not b_type in operations:
                raise ValueError("Binding type %s no in operations" % b_type)
            xb_soap = xb.findall("./{%s}binding" % xmlnamespace.NS_SOAP)
            if not(xb_soap):
                continue  # not a soap binding in wsdl
            if xb_soap[0].get("style", "") == "rpc":
                continue
            if xb_soap[0].get("transport", "") !=\
                    "http://schemas.xmlsoap.org/soap/http":
                continue
            ops = operations[b_type]
            bindings[b_name] = {}
            xops = xb.findall("./{%s}operation" % xmlnamespace.NS_WSDL)
            for xop in xops:
                op_name = xop.get("name", "")
                if not op_name in ops:
                    raise ValueError("operation %s no in operations" %
                                     op_name)
                soap_op = xop.find("./{%s}operation" % xmlnamespace.NS_SOAP)
                s_action = None
                if soap_op is not None:
                    s_action = soap_op.get("soapAction", "")

                all_literal = True
                xop_in = xop.find("./{%s}input" % xmlnamespace.NS_WSDL)
                if xop_in is not None:
                    xop_in_body = xop_in.find("./{%s}body" % xmlnamespace.NS_SOAP)
                    if xop_in_body is None:
                        raise ValueError("No body found for %s" %
                                         (etree.tostring(xop).decode()))
                    if xop_in_body.get("use") != "literal":
                        all_literal = False
                    parts = xop_in_body.get("parts")
                    if parts is None:
                        ops[op_name].input.use_parts = ops[op_name].input.parts
                    else:
                        parts = parts.split(" ")
                        ops[op_name].input.use_parts = []
                        for p in parts:
                            for pp in ops[op_name].input.parts:
                                if pp[0] == p:
                                    ops[op_name].input.use_parts.append(pp)
                                    break

                xop_out = xop.find("./{%s}output" % xmlnamespace.NS_WSDL)
                if xop_out is not None:
                    xop_out_body = xop_out.find("./{%s}body" % xmlnamespace.NS_SOAP)
                    if xop_out_body is None:
                        raise ValueError("No body found for %s" %
                                         (etree.tostring(xop).decode()))
                    if xop_out_body.get("use") != "literal":
                        all_literal = False
                    parts = xop_out_body.get("parts")
                    if parts is None:
                        ops[op_name].output.use_parts = ops[op_name].output.parts
                    else:
                        parts = parts.split(" ")
                        ops[op_name].output.use_parts = []
                        for p in parts:
                            for pp in ops[op_name].output.parts:
                                if pp[0] == p:
                                    ops[op_name].output.use_parts.append(pp)
                                    break
                # rebuild __doc__ after messing with messages
                ops[op_name]._redoc()

                if all_literal:
                    ops[op_name].action = s_action
                    bindings[b_name][op_name] = ops[op_name]
        return bindings

    def get_services(self, bindings):
        """
            Find all services an make final list of
            operations.

            This also sets location to all operations.

            Parameters
            ----------
            bindings : dic from get_bindings.

            Returns
            -------
            out : dict
                Dictionary {service -> {operation name -> method}.
        """

        xservices = self.wsdl.findall('./{%s}service' % xmlnamespace.NS_WSDL)
        services = {}
        for xs in xservices:
            s_name = xs.get("name", "")
            xports = xs.findall('./{%s}port' % xmlnamespace.NS_WSDL)
            for xp in xports:
                b = xp.get("binding", "")
                xaddr = xp.findall('./{%s}address' % xmlnamespace.NS_SOAP)
                if not(xaddr):
                    continue  # no soap 11
                loc = xaddr[0].get("location", "")
                if b in bindings:
                    for k, v in bindings[b].items():
                        v.location = loc
                    services[s_name] = bindings[b]
        return services

    def parse(self):
        """
            Do parsing, return types, services.

            Returns
            -------
            out : (types, services)
        """
        t = self.get_types()
        m = self.get_messages(t)
        op = self.get_operations(m)
        b = self.get_bindings(op)
        s = self.get_services(b)
        return t, s



# xmlnamespace.py - data for namespaces, part of osa.
# Copyright 2013 Sergey Bozhenkov, boz at ipp.mpg.de
# Licensed under LGPLv3 or later, see the COPYING file.

"""
    A list of default namespaces and some simplistic help functions.
"""
# default namespaces
NS_SOAP_ENV = "http://schemas.xmlsoap.org/soap/envelope/"
NS_SOAP_ENC = "http://schemas.xmlsoap.org/soap/encoding/"
NS_SOAP = 'http://schemas.xmlsoap.org/wsdl/soap/'
NS_SOAP12 = 'http://schemas.xmlsoap.org/wsdl/soap12/'
NS_XSI = "http://www.w3.org/2001/XMLSchema-instance"
NS_XSD = "http://www.w3.org/2001/XMLSchema"
NS_WSDL = 'http://schemas.xmlsoap.org/wsdl/'


def get_local_name(full_name):
    """
        Removes namespace part of the name.

        Namespaces can appear in 2 forms:
            {full.namespace.com}name, and
            prefix:name.
        Both cases are handled correctly here.
    """
    full_name = full_name[full_name.find('}')+1:]
    full_name = full_name[full_name.find(':')+1:]
    return full_name


def get_ns(tag):
    """
        Extract namespace.

        This function is opposite to get_local_name, in that
        it returns the first part of the tag: the namespace.

        Parameters
        ----------
        tag : str
            Tag to process.
    """
    p_open = tag.find('{')
    p_close = tag.find('}')
    if p_open != -1 and p_close != -1:
        return tag[p_open+1:p_close]
    else:
        return ''



from setuptools import setup, find_packages
import sys

CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Environment :: Console
Environment :: Web Environment
Intended Audience :: Developers
Intended Audience :: End Users/Desktop
License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)
Natural Language :: English
Operating System :: OS Independent
Programming Language :: Python
Topic :: Internet
Topic :: Internet :: WWW/HTTP
Topic :: Software Development :: Libraries
Topic :: Software Development :: Object Brokering""".split("\n")

setup(
    name='osa',
    version='0.2.3',
    description='Python fast/slim/convenient SOAP/WSDL client.',
    author="Sergey A. Bozhenkov",
    author_email='ba-serge@yandex.ru',
    url="https://github.com/baserge/osa",
    download_url="https://github.com/baserge/osa/archive/master.zip",
    packages=["osa",],
    license="LGPLv3",
    classifiers=CLASSIFIERS,
    )



# xmlparser.py - functions for xml parsing, part of osa.
# Copyright 2013 Sergey Bozhenkov, boz at ipp.mpg.de
# Licensed under LGPLv3 or later, see the COPYING file.

"""
    Help functions for dealing with xml.
"""
import xml.etree.cElementTree as etree
import sys
if sys.version_info[0] < 3:
    from urllib2 import urlopen, HTTPError
else:
    from urllib.request import urlopen, HTTPError

default_attr = ["type", "base", "element", "message", "binding", "ref"]


def parse_qualified(f, attr=None):
    """
        Parse xml from file-like object and make changes to qualified values.

        The standard ElementTree parser does not do consider namespaces
        in attribute values. This is however used in wsdl a lot.
        This function uses iterative parser to find the meaning of the
        ns-prefixes and substitutes them with proper values as {...}value.

        Parameters
        ----------
        f : file-like object
            Input source of xml.
        attr : list
            List of attributes whose value must be considered for
            namespace expansion. If none the default_attr is used.

        Returns
        -------
        root : xml node
            Root of the supplied document.
    """
    if attr is None:
        attr = default_attr

    # iterative parsing is done after an example on
    # http://effbot.org/zone/element-namespaces.htm
    ns_map = []  # stack of defined prefixes and values
    events = ("start", "start-ns", "end-ns")
    # start - begin of a new element
    # start-ns - opening of a new namespace
    # end-ns - closing of a namespace
    root = None

    for event, element in etree.iterparse(f, events):
        if event == "start-ns":
            # in this case element is a pair (prefix, value)
            ns_map.append(element)
        elif event == "end-ns":
            ns_map.pop()
        elif event == "start":
            if root is None:
                root = element
            # check that this element has an attribute of interest
            for a in element.attrib:
                ashort = a.split(":")[-1].split("}")[-1]
                if ashort in attr:
                    # check the attribute value is qualified
                    vlist = element.attrib[a].split(":")
                    if len(vlist) == 2:
                        # search for the namespace prefix
                        for ns in ns_map:
                            if ns[0] == vlist[0]:
                                # got it
                                element.attrib[a] = "{%s}%s" % (ns[1], vlist[1])
    return root


def parse_qualified_from_url(url, attr=None, wsdl_url=None):
    """
        The same as `parse_qualified`, but xml is given by its url.

        URL is either  http://, or a file if that prefix is not present.
    """
    # open page - get a file like object and
    # parse it into xml
    try:
        # opens http://, https://, file://
        page_handler = urlopen(url)
    except (HTTPError, ValueError):
        try:
            # url is something /path/to/file, use open directly
            page_handler = open(url, 'r')
        except IOError:
            page_handler = None
            orig_url = url
            if wsdl_url:
                try:
                    # local file isn't found, try to get remote file with
                    # wsdl_url 'directory' + filename
                    url = wsdl_url.rsplit('/', 1)[0] + '/' + url
                    page_handler = urlopen(url)
                except (HTTPError, ValueError):
                    pass
            if not page_handler:
                raise ValueError("'%s' not found." % orig_url)

    root = parse_qualified(page_handler, attr=attr)
    page_handler.close()
    del page_handler
    return root
